/* Name: tempGuy.c
 * Project: tempGuy based on AVR USB driver
 * Author: Christian Starkjohann
 * Creation Date: 2005-01-16
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: tempGuy.c 472 2008-01-21 18:21:59Z cs $
 */

/*
General Description:
This program controls the tempGuy USB device from the command line.
It must be linked with libusb, a library for accessing the USB bus from
Linux, FreeBSD, Mac OS X and other Unix operating systems. Libusb can be
obtained from http://libusb.sourceforge.net/.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */
#include <unistd.h>

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */
/* Use obdev's generic shared VID/PID pair and follow the rules outlined
 * in firmware/usbdrv/USBID-License.txt.
 */

#define PSCMD_READY           0

#define PSCMD_CONVERT_TEMP    1
#define PSCMD_READ_SCRATCH    2
#define PSCMD_FETCH_SCRATCH   3

#define PSCMD_READ_SERIAL     4
#define PSCMD_FETCH_SERIAL    5

/* These are the vendor specific SETUP commands implemented by our USB device */

static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
char    buffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
        return rval;
    if(buffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)buffer[0] < rval)
        rval = (unsigned char)buffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = buffer[2 * i];
        if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}


/* tempGuy uses the free shared default VID/PID. If you want to see an
 * example device lookup where an individually reserved PID is used, see our
 * RemoteSensor reference implementation.
 */

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL){  /* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }else{
                    errorCode = USB_ERROR_NOTFOUND;
                    /* fprintf(stderr, "seen device from vendor ->%s<-\n", string); */
                    if(strcmp(string, vendorName) == 0){
                        len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                        if(len < 0){
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
                        }else{
                            errorCode = USB_ERROR_NOTFOUND;
                            /* fprintf(stderr, "seen product ->%s<-\n", string); */
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}

// #define PSCMD_ECHO            0
// Send a echo command
/* void cmd_echo(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   int                 nBytes;
   int v, r, i;
   srand(time(NULL));
   v = rand() & 0xffff;
   nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
   if(nBytes < 2){
      if(nBytes < 0)
         fprintf(stderr, "USB error: %s\n", usb_strerror());
      fprintf(stderr, "only %d bytes received in test\n", nBytes);
      fprintf(stderr, "value sent = 0x%x\n", v);
      exit(1);
   }
   printf("sent 2 bytes (0x%02x 0x%02x)\n", v & 0xff, (v & 0xff00)/ 256);

   r = buffer[0] | (buffer[1] << 8);
   if(r != v){
      fprintf(stderr, "data error: received 0x%x instead of 0x%x in test\n", r, v);
      // exit(1);
   }

   printf("Received %d bytes: ", nBytes);
   for (i = 0; i < nBytes; i++)
      printf("0x%02x ", (unsigned int) buffer[i]);
   printf("\n");
   printf("test succeeded\n");
}
 */

// #define PSCMD_CONVERT_TEMP    1
// Send Start conversion command
void cmd_convert_temp(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   usb_control_msg(handle,
                   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                   PSCMD_CONVERT_TEMP,
                   0,
                   0,
                   (char *)buffer,
                   sizeof(buffer),
                   5000);
   printf("Convert temperature command sent\n");
}

// #define PSCMD_READ_SCRATCH    2
// Read scratchpad command
void cmd_read_scratch(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   usb_control_msg(handle,
                   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                   PSCMD_READ_SCRATCH,
                   0,
                   0,
                   (char *)buffer,
                   sizeof(buffer),
                   5000);
   printf("Read scratchpad command sent\n");
}

// #define PSCMD_FETCH_SCRATCH   6
// Fetch the result of a read scratchpad command
void cmd_fetch_scratch(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   int                 nBytes;
   int i;
   double t;
   short n;

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_FETCH_SCRATCH,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   printf("Received %d bytes: ", nBytes);
   for (i = 0; i < nBytes; i++)
      printf("0x%02x ", (unsigned int) buffer[i]);
   printf("\n");

   n = buffer[0] | (((char) buffer[1]) << 8);
   t = (n & ~1)/2 -0.25 + (16.0 - buffer[6])/16.0;

   printf("I think this means: %lf (better: %lf)\n", ((double) n)/2, t);
}



// #define PSCMD_READ_SERIAL     3
// Read Serial
void cmd_read_serial(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   int                 nBytes;
   int i;

   printf("Reading Serial ...\n");

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_READ_SERIAL,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   printf("Received %d bytes: ", nBytes);
   for (i = 0; i < nBytes; i++)
      printf("0x%02x ", (unsigned int) buffer[i]);
   printf("\n");
}

// #define PSCMD_FETCH_SERIAL    4
// Fetch the result of a read serial command
void cmd_fetch_serial(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   int                 nBytes;
   int i;

   printf("Reading Serial ...\n");

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_FETCH_SERIAL,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   printf("Received %d bytes: ", nBytes);
   for (i = 0; i < nBytes; i++)
      printf("0x%02x ", (unsigned int) buffer[i]);
   printf("\n");
}

// #define PSCMD_READY           5
// Check wether the chip is ready to accept a new command
unsigned char cmd_ready(usb_dev_handle      *handle)
{
   unsigned char       buffer[8];
   int                 nBytes;
   int i;

   printf("Reading Ready Flag ...\n");

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_READY,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   printf("Received %d bytes: ", nBytes);
   for (i = 0; i < nBytes; i++)
      printf("0x%02x ", (unsigned int) buffer[i]);
   printf("\n");

   return 1 == nBytes && buffer[0];
}

int main(int argc, char **argv)
{
usb_dev_handle      *handle = NULL;

    usb_init();
    if(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "hoepfl.de", USBDEV_SHARED_PRODUCT, "TempGuy") != 0){
        fprintf(stderr, "Could not find USB device \"TempGuy\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
        exit(1);
    }

    // cmd_echo(handle);

    cmd_read_serial(handle);
    // sleep(1);
    usleep(10000);
    while (!cmd_ready(handle))
       usleep(10000);
    cmd_fetch_serial(handle);


    while(1)
    {
       cmd_convert_temp(handle);
       sleep(1);
       while (!cmd_ready(handle))
          usleep(10000);
       usleep(10000);
       cmd_read_scratch(handle);
       usleep(10000);
       while (!cmd_ready(handle))
          usleep(10000);
       cmd_fetch_scratch(handle);
    }

    usb_close(handle);
    return 0;
}
