#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <usb.h>
#include <unistd.h>

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */

// Commands known by TempGuy
#define PSCMD_READY           0

#define PSCMD_CONVERT_TEMP    1
#define PSCMD_READ_SCRATCH    2
#define PSCMD_FETCH_SCRATCH   3

#define PSCMD_READ_SERIAL     4
#define PSCMD_FETCH_SERIAL    5

static int usbGetStringAscii(usb_dev_handle *dev,
                             int index,
                             int langid,
                             char *buf,
                             int buflen)
{
   char buffer[256];
   int rval, i;

   if((rval = usb_control_msg(dev,
                              USB_ENDPOINT_IN,
                              USB_REQ_GET_DESCRIPTOR,
                              (USB_DT_STRING << 8) + index,
                              langid,
                              buffer,
                              sizeof(buffer),
                              1000)) < 0)
   {
      return rval;
   }

   if(buffer[1] != USB_DT_STRING)
   {
      return 0;
   }
   if((unsigned char)buffer[0] < rval)
   {
      rval = (unsigned char)buffer[0];
   }
   rval /= 2;

   // lossy conversion to ISO Latin1
   for(i=1;i<rval;i++) {
      if(i > buflen)  /* destination buffer overflow */
         break;
      buf[i-1] = buffer[2 * i];
      if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
         buf[i-1] = '?';
   }
   buf[i-1] = 0;

   return i-1;
}

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device,
                         int vendor,
                         char *vendorName,
                         int product,
                         char *productName)
{
   struct usb_bus    *bus;
   struct usb_device *dev;
   usb_dev_handle    *handle = NULL;
   int               errorCode = USB_ERROR_NOTFOUND;
   static int        didUsbInit = 0;
   
   if(!didUsbInit)
   {
      didUsbInit = 1;
      usb_init();
   }

   usb_find_busses();
   usb_find_devices();

   for(bus=usb_get_busses(); bus; bus=bus->next)
   {
      for(dev=bus->devices; dev; dev=dev->next)
      {
         if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product)
         {
            char  string[256];
            int   len;

            handle = usb_open(dev); /* we need to open the device in order to query strings */
            if(!handle)
            {
               errorCode = USB_ERROR_ACCESS;
               fprintf(stderr,
                       "Warning: cannot open USB device: %s\n",
                       usb_strerror());
               continue;
            }
            if(vendorName == NULL && productName == NULL)
            {  /* name does not matter */
               break;
            }
   
            // now check whether the names match:
            len = usbGetStringAscii(handle,
                                    dev->descriptor.iManufacturer,
                                    0x0409,
                                    string,
                                    sizeof(string));
            if(len < 0)
            {
               errorCode = USB_ERROR_IO;
               fprintf(stderr,
                       "Warning: cannot query manufacturer for device: %s\n",
                       usb_strerror());
            }
            else
            {
               /* fprintf(stderr, "seen device from vendor ->%s<-\n", string); */

               errorCode = USB_ERROR_NOTFOUND;
               if(strcmp(string, vendorName) == 0)
               {
                  len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                  if(len < 0)
                  {
                     errorCode = USB_ERROR_IO;
                     fprintf(stderr,
                             "Warning: cannot query product for device: %s\n",
                             usb_strerror());
                  }
                  else
                  {
                     /* fprintf(stderr, "seen product ->%s<-\n", string); */

                     errorCode = USB_ERROR_NOTFOUND;
                     if (strcmp(string, productName) == 0)
                        break;
                  }
               }
            }
            usb_close(handle);
            handle = NULL;
         }
      }
      if(handle)
         break;
   }
   if(handle != NULL)
   {
      errorCode = 0;
      *device = handle;
   }
   return errorCode;
}

// Send Start conversion command
void cmd_convert_temp(usb_dev_handle *handle)
{
   unsigned char buffer[8];
   usb_control_msg(handle,
                   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                   PSCMD_CONVERT_TEMP,
                   0,
                   0,
                   (char *)buffer,
                   sizeof(buffer),
                   5000);
   /* printf("Convert temperature command sent\n"); */
   sleep(1);
}

// Read scratchpad command
void cmd_read_scratch(usb_dev_handle *handle)
{
   unsigned char buffer[8];
   usb_control_msg(handle,
                   USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                   PSCMD_READ_SCRATCH,
                   0,
                   0,
                   (char *)buffer,
                   sizeof(buffer),
                   5000);
   /* printf("Read scratchpad command sent\n"); */
}

// Fetch the result of a read scratchpad command
void cmd_fetch_scratch(usb_dev_handle *handle)
{
   unsigned char  buffer[8];
   int            nBytes;
   double         t;
   short          n;

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_FETCH_SCRATCH,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   /* {
      int i;
      printf("Received %d bytes: ", nBytes);
      for (i = 0; i < nBytes; i++)
         printf("0x%02x ", (unsigned int) buffer[i]);
      printf("\n");
   } */

   n = buffer[0] | (((char) buffer[1]) << 8);
   t = (n & ~1)/2 - 0.25 + (16.0 - buffer[6])/16.0;

   printf("%.1lf %.4lf\n", ((double) n)/2, t);
}

// Read Serial
void cmd_read_serial(usb_dev_handle *handle)
{
   unsigned char buffer[8];
   int nBytes;

   /* printf("Reading Serial ...\n"); */

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_READ_SERIAL,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   /* {
      int i;
      printf("Received %d bytes: ", nBytes);
      for (i = 0; i < nBytes; i++)
         printf("0x%02x ", (unsigned int) buffer[i]);
      printf("\n");
   } */
}

// Fetch the result of a read serial command
void cmd_fetch_serial(usb_dev_handle *handle)
{
   unsigned char  buffer[8];
   int            nBytes;

   /* printf("Reading Serial ...\n"); */

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_FETCH_SERIAL,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   /* {
      int i;
      printf("Received %d bytes: ", nBytes);
      for (i = 0; i < nBytes; i++)
         printf("0x%02x ", (unsigned int) buffer[i]);
      printf("\n");
   } */
}

// Check wether the chip is ready to accept a new command
unsigned char cmd_ready(usb_dev_handle *handle)
{
   unsigned char  buffer[8];
   int            nBytes;

   /* printf("Reading Ready Flag ...\n"); */

   nBytes = usb_control_msg(handle,
                            USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                            PSCMD_READY,
                            0,
                            0,
                            (char *)buffer,
                            sizeof(buffer),
                            5000);

   /* {
      int i;
      printf("Received %d bytes: ", nBytes);
      for (i = 0; i < nBytes; i++)
         printf("0x%02x ", (unsigned int) buffer[i]);
      printf("\n");
   } */

   return 1 == nBytes && buffer[0];
}

int main(int argc, char *argv[])
{
   usb_dev_handle *handle = NULL;

   usb_init();
   if(usbOpenDevice(&handle,
                    USBDEV_SHARED_VENDOR,
                    "hoepfl.de",
                    USBDEV_SHARED_PRODUCT,
                    "TempGuy") != 0)
   {
      fprintf(stderr,
              "Could not find USB device \"TempGuy\" with vid=0x%x pid=0x%x\n",
              USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
      exit(1);
   }

   cmd_read_serial(handle);
   do { usleep(10000); } while (!cmd_ready(handle));
   cmd_fetch_serial(handle);

   cmd_convert_temp(handle);
   do { usleep(10000); } while (!cmd_ready(handle));
   cmd_read_scratch(handle);
   do { usleep(10000); } while (!cmd_ready(handle));
   cmd_fetch_scratch(handle);

   usb_close(handle);
   return 0;
}
