#include "DS18S20.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#ifndef USB_CONCAT
#define USB_CONCAT(a, b)            a ## b
#define USB_CONCAT_EXPANDED(a, b)   USB_CONCAT(a, b)
#endif

#define DS_OUTPORT USB_CONCAT_EXPANDED(PORT, DS_PORT)
#define DS_INPORT  USB_CONCAT_EXPANDED(PIN, DS_PORT)
#define DS_DDR     USB_CONCAT_EXPANDED(DDR, DS_PORT)
#define DS_MASK    (1 << DS_BIT)

static DS_RETURNTYPE ds_initialization(void)
{
#if DS_RETURNRESULTS
   unsigned char ok = 0;
#endif

   // Pull line to low
   DS_OUTPORT &= ~DS_MASK;
   DS_DDR |= DS_MASK;

   // Hold line low for >480us
   _delay_us(480);

   // Disable interrupts
   cli();

   // Revert to input mode
   DS_DDR &= ~DS_MASK;

   // Give DS1820 some time to set its precence signal
   _delay_us(60);

#if DS_RETURNRESULTS
   // Read precence signal
   ok = DS_INPORT & DS_MASK;
#endif

   // Reenable interrupts
   // We did not enable them above since interrupts could make us
   // miss the precence signal.
   sei();

#if DS_RETURNRESULTS
   // The line must be pulled to low here
   if (0 == ok)
   {
#endif

      // We read again at the end of 480us
      _delay_us(420);

#if DS_RETURNRESULTS
      // The presence signal must not be there any longer
      if (DS_INPORT & DS_MASK)
      {
         ok = 1;
      }
   }

   return ok;
#endif
}

static void ds_write_1(void)
{
   // Pull line to low, disable interrupts
   DS_OUTPORT &= ~DS_MASK;
   cli();
   DS_DDR |= DS_MASK;

   // Hold line low for 1us
   _delay_us(1);

   // Pull line to high, revert to input mode
   // (line is still pulled to high by external pull-up)
   // enable interrupts.
   DS_OUTPORT |= DS_MASK;
   DS_DDR &= ~DS_MASK;
   sei();
   DS_OUTPORT &= ~DS_MASK;

   // make sure the line is not used for 59us + 1us gap
   _delay_us(60);
}

static void ds_write_0(void)
{
   // Pull line to low, disable interrupts
   DS_OUTPORT &= ~DS_MASK;
   cli();
   DS_DDR |= DS_MASK;

   // Hold line low for 60us
   _delay_us(60);

   // Pull line to high, revert to input mode
   // (line is still pulled to high by external pull-up)
   // enable interrupts
   DS_OUTPORT |= DS_MASK;
   DS_DDR &= ~DS_MASK;
   sei();
   DS_OUTPORT &= ~DS_MASK;

   // make sure the line is not used for 1us
   _delay_us(1);
}

static unsigned char ds_read_bit(void)
{
   unsigned char result;

   // Pull line to low, disallow interrupts
   DS_OUTPORT &= ~DS_MASK;
   cli();
   DS_DDR |= DS_MASK;

   // Hold low for 1us
   _delay_us(1);

   // Return to input mode (this pulls line to high)
   DS_DDR &= ~DS_MASK;

   // Give DS1820 14us time to answer
   _delay_us(14);

   // Read bit
   result = DS_INPORT & DS_MASK;

   // Reenable interrupts
   sei();

   // Wait for DS1820 to release line
   while (!(DS_INPORT & DS_MASK))
      ;

   // Wait 45us (zero read) + 1us as inter bit gap
   _delay_us(46);

   return result;
}

static void ds_sendByte(unsigned char value)
{
   int i;
   for (i = 8; i; --i)
   {
      if (value & 1)
         ds_write_1();
      else
         ds_write_0();

      value >>= 1;
   }
}

static unsigned char ds_readByte()
{
   unsigned char b;
   unsigned char result = 0;

   for (b = 0x01; b; b <<= 1)
   {
      if (ds_read_bit())
      {
         result |= b;
      }
   }

   return result;
}



DS_RETURNTYPE ds_convert(void)
{
#if DS_RETURNRESULTS
   unsigned char ok;

   ok = ds_initialization();
   if (ok)
   {
#else
   ds_initialization();
#endif
      ds_sendByte(0xcc);
      ds_sendByte(0x44);
#if DS_RETURNRESULTS 
   }

   return ok;
#endif
}

DS_RETURNTYPE ds_readScratchpad(unsigned char *buffer)
{
#if DS_RETURNRESULTS
   unsigned char ok;
#endif
   int b;

#if DS_RETURNRESULTS
   ok = ds_initialization();
   if (ok)
   {
#else
      ds_initialization();
#endif

      ds_sendByte(0xcc);
      ds_sendByte(0xbe);

      for (b = 0; b < 9; b++)
      {
         buffer[b] = ds_readByte();
      }
#if DS_RETURNRESULTS
   }

   return ok;
#endif
}

DS_RETURNTYPE ds_readSerial(unsigned char *buffer)
{
   int b;

#if DS_RETURNRESULTS
   unsigned char ok;

   ok = ds_initialization();
   if (ok)
   {
#else
   ds_initialization();
#endif
      ds_sendByte(0x33);

      for (b = 0; b < 8; b++)
      {
         buffer[b] = ds_readByte();
      }

      ds_sendByte(0x00);
#if DS_RETURNRESULTS 
   }

   return ok;
#endif
}
