/* Name: main.c
 * Project: TempGuy based on AVR USB driver
 * Author: Daniel Hoepfl
 * Creation Date: 2008-09-24
 * Tabsize: 4
 * Copyright: (c) 2008 by Daniel Hoepfl
 * License: GNU GPL v2
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "DS18S20.h"
#include <util/delay.h>

#if FLASHEND <= 0x7ff
#warning *** No GET_SERIAL command ***
#warning *** No Watchdog ***
#endif

static uchar   sReadyBuf[1];
static uchar   sResultBuffer[9];

typedef void (*idleFunctionT)(void);
volatile idleFunctionT idleFunction = NULL;

void idleConvert(void) { ds_convert(); }
void idleSerial(void)  { ds_readSerial(sResultBuffer); }
void idleScratch(void) { ds_readScratchpad(sResultBuffer); }

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET.
 */
void hadUsbReset(void)
{
    const int targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);
    uchar       step = 128;
    uchar       trialValue = 0;
    int diff;
    int mdiff, pdiff;

    /* do a binary search: */
    do {
        OSCCAL = trialValue + step;
        diff = targetValue - usbMeasureFrameLength();
        if (diff >= 0)
        {
            trialValue += step;
        }
        else
        {
            diff = -diff;
        }
        step >>= 1;
    } while(step > 0);

    // Find best value (+/-0 is already known)
    OSCCAL++;
    pdiff = targetValue - usbMeasureFrameLength();
    OSCCAL -= 2;
    mdiff = usbMeasureFrameLength() - targetValue;

    // Is diff better than pdiff? (This means diff or mdiff is best)
    if (diff < pdiff)
    {
        // Is diff the best? If not: mdiff is already active.
        if (diff < mdiff)
        {
            OSCCAL = trialValue;
        }
    }
    else
    {
        // diff wasn't the best match, choose between mdiff an pdiff.
        // If mdiff is the best match: do nothing since it is already active.
        if (pdiff < mdiff)
        {
            // pdiff was the best match.
            OSCCAL = trialValue+1;
        }
    }
} 

USB_PUBLIC uchar usbFunctionSetup(uchar data[8])
{
   usbRequest_t    *rq = (void *)data;

   if (0 == rq->bRequest) /* Ready for a new command? */
   {
      sReadyBuf[0] = idleFunction == NULL;

      usbMsgPtr = sReadyBuf;
      return 1;
   }
   else if (1 == rq->bRequest) /* Start convert */
   {
      idleFunction = idleConvert;
   }
   else if (2 == rq->bRequest) /* Get Temperature */
   {
      idleFunction = idleScratch;
   }
   else if (3 == rq->bRequest) /* Get Temperature result */
   {
      usbMsgPtr = sResultBuffer;
      return 8;
   }
#if FLASHEND > 0x7ff	// Include in 4/8k chips only
   else if (4 == rq->bRequest) /* Get serial */
   {
      idleFunction = idleSerial;
   }
   else if (5 == rq->bRequest) /* Get Serial Result */
   {
      usbMsgPtr = sResultBuffer;
      return 8;
   }
#endif

   return 0;
}

int main(void)
{
   uchar   i;

#if FLASHEND > 0x7ff	// Include in 4/8k chips only
   wdt_enable(WDTO_1S);
#endif

   USBOUT = 0;          /* no pullups on USB pins */
   USBDDR = ~USBMASK;    /* set all pins as outputs except USB */
   USBOUT = 0;          /* no pullups on USB pins */
   /* We fake an USB disconnect by pulling D+ and D- to 0 during reset. This is
    * necessary if we had a watchdog reset or brownout reset to notify the host
    * that it should re-enumerate the device. Otherwise the host's and device's
    * concept of the device-ID would be out of sync.
    */
   usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
   i = 0;
   while(--i){         /* fake USB disconnect for > 500 ms */
#if FLASHEND > 0x7ff	// Include in 4/8k chips only
      wdt_reset();
#endif
      _delay_ms(2);
   }
   usbDeviceConnect();
   usbInit();

   sei();
   for(;;){    /* main event loop */
#if FLASHEND > 0x7ff	// Include in 4/8k chips only
      wdt_reset();
#endif

      usbPoll();

      if (idleFunction)
      {
         i = 0;
         uchar oldCount = usbSofCount;
         while (--i && oldCount == usbSofCount)
             ;

         if (oldCount != usbSofCount)
         {
            idleFunction();
            idleFunction = NULL;
         }
      }

   }
   return 0;
}

